#include <stdbool.h>
#include <stdio.h>
#include <talloc.h>
#define ARRAY_SIZE(arr) (sizeof (arr) / sizeof (arr[0]))

/* Structure printer interface */
struct sprinter
{
    void (*begin_map) (struct sprinter *);
    void (*begin_list) (struct sprinter *);
    void (*end) (struct sprinter *);

    void (*string) (struct sprinter *, const char *);
    void (*integer) (struct sprinter *, int);
    void (*boolean) (struct sprinter *, bool);
    void (*null) (struct sprinter *);
    void (*map_key) (struct sprinter *, const char *);

    void (*frame) (struct sprinter *);
};

/* Create a new structure printer that emits JSON */
struct sprinter *
sprinter_json_new(const void *ctx, FILE *stream);

int
main(int argc, char **argv)
{
    struct sprinter *test = sprinter_json_new (NULL, stdout);
    test->begin_list (test);
    test->string (test, "test");
    test->integer (test, 42);
    test->boolean (test, true);
    test->null (test);
    test->frame (test);
    test->begin_map (test);
    test->map_key (test, "Hello");
    test->string (test, "world\n");
    test->end (test);
    test->frame (test);
    test->end (test);
    talloc_free (test);
}

/*
 * Every below here is private implementation.
 */

struct sprinter_json
{
    struct sprinter vtable;
    FILE *stream;
    /* Top of the state stack, or NULL if the printer is not currently
     * inside any aggregate types. */
    struct json_state *state;
};

struct json_state
{
    struct json_state *parent;
    /* True if nothing has been printed in this aggregate yet.
     * Suppresses the comma before a value. */
    bool first;
    /* The character that closes the current aggregate. */
    char close;
};

/* Helper function to set up the stream to print a value.  If this
 * value follows another value, prints a comma. */
static struct sprinter_json *
json_begin_value(struct sprinter *sp)
{
    struct sprinter_json *spj = (struct sprinter_json*)sp;
    if (spj->state) {
	if (!spj->state->first)
	    fputs (", ", spj->stream);
	else
	    spj->state->first = false;
    }
    return spj;
}

/* Helper function to begin an aggregate type.  Prints the open
 * character and pushes a new state frame. */
static void
json_begin_aggregate(struct sprinter *sp, char open, char close)
{
    struct sprinter_json *spj = json_begin_value (sp);
    struct json_state *state = talloc (spj, struct json_state);

    fputc (open, spj->stream);
    state->parent = spj->state;
    state->first = true;
    state->close = close;
    spj->state = state;
}

static void
json_begin_map(struct sprinter *sp)
{
    json_begin_aggregate (sp, '{', '}');
}

static void
json_begin_list(struct sprinter *sp)
{
    json_begin_aggregate (sp, '[', ']');
}

static void
json_end(struct sprinter *sp)
{
    struct sprinter_json *spj = (struct sprinter_json*)sp;
    struct json_state *state = spj->state;

    fputc (spj->state->close, spj->stream);
    spj->state = state->parent;
    talloc_free (state);
    if(spj->state == NULL)
	fputc ('\n', spj->stream);
}

static void
json_string(struct sprinter *sp, const char *val)
{
    const static char * const escapes[] = {
	['\"'] = "\\\"", ['\\'] = "\\\\", ['\b'] = "\\b",
	['\f'] = "\\f",  ['\n'] = "\\n",  ['\t'] = "\\t"
    };
    struct sprinter_json *spj = json_begin_value (sp);
    fputc ('"', spj->stream);
    for (; *val; ++val) {
	unsigned char ch = *val;
	if (ch < ARRAY_SIZE(escapes) && escapes[ch])
	    fputs (escapes[ch], spj->stream);
	else if (ch >= 32)
	    fputc (ch, spj->stream);
	else
	    fprintf (spj->stream, "\\u%04x", ch);
    }
    fputc ('"', spj->stream);
}

static void
json_integer(struct sprinter *sp, int val)
{
    struct sprinter_json *spj = json_begin_value (sp);
    fprintf (spj->stream, "%d", val);
}

static void
json_boolean(struct sprinter *sp, bool val)
{
    struct sprinter_json *spj = json_begin_value (sp);
    fputs (val ? "true" : "false", spj->stream);
}

static void
json_null(struct sprinter *sp)
{
    struct sprinter_json *spj = json_begin_value (sp);
    fputs ("null", spj->stream);
}

static void
json_map_key(struct sprinter *sp, const char *key)
{
    struct sprinter_json *spj = (struct sprinter_json*)sp;
    json_string (sp, key);
    fputs (": ", spj->stream);
    spj->state->first = true;
}

static void
json_frame(struct sprinter *sp)
{
    struct sprinter_json *spj = (struct sprinter_json*)sp;
    fputc ('\n', spj->stream);
}

struct sprinter *
sprinter_json_new(const void *ctx, FILE *stream)
{
    const static struct sprinter_json template = {
	.vtable = {
	    .begin_map = json_begin_map,
	    .begin_list = json_begin_list,
	    .end = json_end,
	    .string = json_string,
	    .integer = json_integer,
	    .boolean = json_boolean,
	    .null = json_null,
	    .map_key = json_map_key,
	    .frame = json_frame,
	}
    };
    struct sprinter_json *res;

    res = talloc (ctx, struct sprinter_json);
    if (!res)
	return NULL;

    *res = template;
    res->stream = stream;
    return &res->vtable;
}
